@@ -216,6 +216,9 @@ Entity NewExternalEntityN(const Char *name, int namelen, const char8 *publicid,
 
     e->url = 0;
 
+    e->is_externally_declared = 0;
+    e->is_internal_subset = 0;
+
     return (Entity)e;
 }
 
@@ -246,6 +249,9 @@ Entity NewInternalEntityN(const Char *name, int namelen,
 
     e->url = 0;
 
+    e->is_externally_declared = 0;
+    e->is_internal_subset = 0;
+
     /* These should not really be accessed for an internal entity, but
        because the NSL layer uses internal entities to implement reading
        from strings, sometimes they are. */
@@ -389,6 +395,11 @@ Entity NextEntity(Dtd dtd, Entity previous)
     return previous ? previous->next : dtd->entities;
 }
 
+Entity NextParameterEntity(Dtd dtd, Entity previous)
+{
+    return previous ? previous->next : dtd->parameter_entities;
+}
+
 /* Elements */
 
 /* 
@@ -451,8 +462,10 @@ ElementDefinition DefineElementN(Dtd dtd, const Char *name, int namelen,
 	return 0;
     e->id_attribute = 0;
     e->xml_space_attribute = 0;
+    e->xml_lang_attribute = 0;
     e->notation_attribute = 0;
     e->cached_nsdef = 0;
+    e->is_externally_declared = 0;
 
     if((t = Strchr(name, ':')))
     {
@@ -611,8 +624,9 @@ AttributeDefinition
     NSL_Doctype_I *doctype = element->doctype;
 #endif
     AttributeDefinition a;
-    static Char xml_space[] = {'x','m','l',':','s','p','a','c','e',0};
-    static Char xmlns[] = {'x','m','l','n','s',0};
+    static Char xml_space[] = {'x','m','l',':','s','p','a','c','e',0}; 
+    static Char xml_lang[] = {'x','m','l',':','l','a','n','g',0};
+   static Char xmlns[] = {'x','m','l','n','s',0};
     Char *t;
 
     if(!(a= Malloc(sizeof(*a))))
@@ -656,6 +670,7 @@ AttributeDefinition
     a->allowed_values = allowed_values;
     a->default_type = default_type;
     a->default_value = default_value;
+    a->is_externally_declared = 0;
 
     if(a->type == AT_id && !element->id_attribute)
 	element->id_attribute = a;
@@ -664,6 +679,8 @@ AttributeDefinition
 
     if(Strcmp(name, xml_space) == 0)
 	element->xml_space_attribute = a;
+    else if(Strcmp(name, xml_lang) == 0)
+	element->xml_lang_attribute = a;
 
     a->cached_nsdef = 0;
 
@@ -804,7 +821,8 @@ void FreeAttributeDefinition(AttributeDefinition a)
  */
 
 NotationDefinition DefineNotationN(Dtd dtd, const Char *name, int namelen, 
-				  const char8 *publicid, const char8 *systemid)
+				  const char8 *publicid, const char8 *systemid,
+				   Entity parent)
 {
     NotationDefinition n;
 
@@ -815,6 +833,8 @@ NotationDefinition DefineNotationN(Dtd dtd, const Char *name, int namelen,
     n->tentative = 0;
     n->systemid = systemid;
     n->publicid = publicid;
+    n->url = 0;
+    n->parent = parent;
     n->next = dtd->notations;
     dtd->notations = n;
 
@@ -826,7 +846,7 @@ NotationDefinition TentativelyDefineNotationN(Dtd dtd,
 {
     NotationDefinition n;
 
-    if(!(n = DefineNotationN(dtd, name, namelen, 0, 0)))
+    if(!(n = DefineNotationN(dtd, name, namelen, 0, 0, 0)))
 	return 0;
 
     n->tentative = 1;
@@ -835,11 +855,13 @@ NotationDefinition TentativelyDefineNotationN(Dtd dtd,
 }
 
 NotationDefinition RedefineNotation(NotationDefinition n,
-				  const char8 *publicid, const char8 *systemid)
+				  const char8 *publicid, const char8 *systemid,
+				    Entity parent)
 {
     n->tentative = 0;
     n->systemid = systemid;
     n->publicid = publicid;
+    n->parent = parent;
 
     return n;
 }
@@ -855,6 +877,23 @@ NotationDefinition FindNotationN(Dtd dtd, const Char *name, int namelen)
     return 0;
 }
 
+NotationDefinition NextNotationDefinition(Dtd dtd, NotationDefinition previous)
+{
+    return previous ? previous->next : dtd->notations;
+}
+
+const char8 *NotationURL(NotationDefinition n)
+{
+    if(n->url)
+	return n->url;
+
+    n->url = url_merge(n->systemid, 
+		       n->parent ? EntityBaseURL(n->parent) : 0, 
+		       0, 0, 0, 0);
+
+    return n->url;
+}
+
 void FreeNotationDefinition(NotationDefinition n)
 {
     if(!n)
@@ -868,6 +907,7 @@ void FreeNotationDefinition(NotationDefinition n)
 
     Free((void *)n->systemid);
     Free((void *)n->publicid);
+    Free((void *)n->url);
 
     /* Free the notation definition itself */
 
